/* arbol_AVL.h */

#include <stdio.h>
#include <stdlib.h>

typedef int tipo_clave;

typedef struct nodo {
  tipo_clave clave;
  struct nodo *izq;
  struct nodo *der;
  int equil; /* -1, 0, 1 */      
} nodo_AVL;

/* puntero a nodo si correcto. NULL en otro caso */
nodo_AVL *crear_nodo(tipo_clave);

/* puntero a nodo si encontrado. NULL en otro caso */
nodo_AVL *buscar_clave(nodo_AVL *, tipo_clave);

/* Devuelve 1 si insertado en hijo y hay que reequilibrar. 0 en otro caso. -3 error */
int insertar_clave_recur(nodo_AVL **, tipo_clave);

/* Devuelve equilibrio de nodo hijo */
int borrar_clave_recur(nodo_AVL **, tipo_clave);

void imprime_inorden_inverso(nodo_AVL *, int);

/* arbol_AVL.c */
nodo_AVL *crear_nodo(tipo_clave entrada) {
nodo_AVL *punt;
  punt = malloc (sizeof(nodo_AVL));
  if (punt != NULL) {
    punt->clave = entrada;
    punt->izq = punt->der = NULL;
    punt->equil = 0;
  } /* if ... */
  return punt;          
} /* crear_nodo */

/* Simple */
int rotacion_II(nodo_AVL **nodo) {
nodo_AVL *nodo1 = (*nodo)->izq;
int dev = 0;

  (*nodo)->izq = nodo1->der;
  nodo1->der = *nodo;

  if (nodo1->equil == -1) {
    nodo1->equil = 0;
    (*nodo)->equil = 0;
    dev = -1; /* utilizado para borrado */    
  } /* if */
  else /* nodo1->equil == 0 */ { /* utilizado para borrado */
    nodo1->equil = 1;
    (*nodo)->equil = -1;
  } /* else */
  
  (*nodo) = nodo1;
  return dev;  
} /* rotacion_II */

/* Simple */
int rotacion_DD(nodo_AVL **nodo) {
nodo_AVL *nodo1 = (*nodo)->der;
int dev = 0;

  (*nodo)->der = nodo1->izq;
  nodo1->izq = *nodo;

  if (nodo1->equil == 1) {
    nodo1->equil = 0;
    (*nodo)->equil = 0;
    dev = -1; /* utilizado para borrado */    
  } /* if */
  else /* nodo1->equil == 0 */ { /* utilizado para borrado */
    nodo1->equil = -1;
    (*nodo)->equil = 1;
  } /* else */
  
  (*nodo) = nodo1;
  return dev;       
} /* rotacion_DD */

/* Doble */
int rotacion_ID(nodo_AVL **nodo) {
nodo_AVL *nodo1 = (*nodo)->izq;
nodo_AVL *nodo2 = nodo1->der;
int dev = 0;

  (*nodo)->izq = nodo2->der;
  nodo2->der = (*nodo);
  nodo1->der = nodo2->izq;
  nodo2->izq = nodo1;

  if (nodo2->equil == -1) {
    nodo1->equil = 0;
    (*nodo)->equil = 1;
  } /* if */
  else if (nodo2->equil == 1) {
    nodo1->equil = -1;
    (*nodo)->equil = 0;   
  } /* else if */
  else /* nodo2->equil == 0 */ { /* utilizado para borrado */
    nodo1->equil = 0;
    (*nodo)->equil = 0;     
  } /* else */
  nodo2->equil = 0;
  
  (*nodo) = nodo2;
  return dev;      
} /* rotacion_ID */

/* Doble */
int rotacion_DI(nodo_AVL **nodo) {
nodo_AVL *nodo1 = (*nodo)->der;
nodo_AVL *nodo2 = nodo1->izq;
int dev = 0;

  (*nodo)->der = nodo2->izq;
  nodo2->izq = (*nodo);
  nodo1->izq = nodo2->der;
  nodo2->der = nodo1;

  if (nodo2->equil == -1) {
    nodo1->equil = 1;
    (*nodo)->equil = 0;
  } /* if */
  else if (nodo2->equil == 1) {
    nodo1->equil = 0;
    (*nodo)->equil = -1;
  } /* else if */
  else /* nodo2->equil == 0 */ { /* utilizado para borrado */
    nodo1->equil = 0;
    (*nodo)->equil = 0;     
  } /* else */  
  nodo2->equil = 0;
  
  (*nodo) = nodo2;
  return dev;  
} /* rotacion_DI */

nodo_AVL *buscar_clave(nodo_AVL *punt_a_buscar, tipo_clave a_buscar) {

  while (punt_a_buscar != NULL) { /* No hemos llegado al final del rbol (hoja) */
    if (punt_a_buscar->clave == a_buscar)
      break; /* equivale a return punt_a_buscar; */
    else if (punt_a_buscar->clave < a_buscar)
      punt_a_buscar = punt_a_buscar->der;
    else /* (punt_a_buscar->clave > a_buscar) */
      punt_a_buscar = punt_a_buscar->izq;  
  } /* while */
  return punt_a_buscar; /* Si encontrado, puntero a clave. En caso contrario NULL */      
} /* buscar_clave */

int insertar_clave_recur(nodo_AVL **punt_a_buscar, tipo_clave a_insertar) {
int dev;
  if (*punt_a_buscar == NULL) {
    *punt_a_buscar = crear_nodo(a_insertar);
    return 1;
  } /* if */
  else
    if ((*punt_a_buscar)->clave == a_insertar) /* error clave repetida */
      return -3;
    else if ((*punt_a_buscar)->clave < a_insertar) {
      dev = insertar_clave_recur(&((*punt_a_buscar)->der), a_insertar);
      if (dev == -3)
        return dev;
        
      if (dev == 1) {
        switch ((*punt_a_buscar)->equil) {
          case -1: (*punt_a_buscar)->equil = 0;
                   break;
          case 0: (*punt_a_buscar)->equil = 1;
                  return 1;
          case 1: if (((*punt_a_buscar)->der)->equil == -1)
                    dev = rotacion_DI(punt_a_buscar);
                  else
                    dev = rotacion_DD(punt_a_buscar);
                  break;    
        } /* switch*/       
      } /* if */
      return 0;
    } /* else if */
    else /* ((*punt_a_buscar)->clave > a_insertar) */ {
      dev = insertar_clave_recur(&((*punt_a_buscar)->izq), a_insertar);
      if (dev == -3)
        return dev;
      
      if (dev == 1) {
        switch ((*punt_a_buscar)->equil) {
          case 1: (*punt_a_buscar)->equil = 0;
                  break;;
          case 0: (*punt_a_buscar)->equil = -1;
                  return 1;
          case -1: if (((*punt_a_buscar)->izq)->equil == 1)
                     dev = rotacion_ID(punt_a_buscar);
                   else
                     dev = rotacion_II(punt_a_buscar);
                   break;    
        } /* switch*/       
      } /* if */
      return 0;
    } /* else */
} /* insertar_clave_recur */

int reemplazar(nodo_AVL **punt_a_buscar, nodo_AVL **punt_a_liberar) {
int dev;
  if ((*punt_a_buscar)->izq == NULL) {
    (*punt_a_liberar)->clave = (*punt_a_buscar)->clave;
    *punt_a_liberar = *punt_a_buscar;
    *punt_a_buscar = (*punt_a_buscar)->der;
    return -1;
  } /* if */
  else {
    dev = reemplazar(&((*punt_a_buscar)->izq), punt_a_liberar);
    
    if (dev == -1) {
      switch ((*punt_a_buscar)->equil) {
        case -1: (*punt_a_buscar)->equil = 0;
                 return -1;
        case 0: (*punt_a_buscar)->equil = 1;
                return 0;
        case 1: if (((*punt_a_buscar)->der)->equil == -1) {
                  dev = rotacion_DI(punt_a_buscar);
                  dev = -1;
                } /* if */
                else
                  dev = rotacion_DD(punt_a_buscar);
                return dev;    
        } /* switch*/       
      } /* if */
      return 0;    
  } /* else */ 
} /* reemplazar */

int borrar_clave_recur(nodo_AVL **punt_a_buscar, tipo_clave a_borrar) {
nodo_AVL *punt_aux;
int dev;
    
  if (*punt_a_buscar == NULL) /* error clave no encontrada */
    return -3;
  
  if ((*punt_a_buscar)->clave < a_borrar) {
    dev = borrar_clave_recur(&((*punt_a_buscar)->der), a_borrar);

    if (dev == -3)
      return dev;
      
    if (dev == -1) {
      switch ((*punt_a_buscar)->equil) {
        case 1: (*punt_a_buscar)->equil = 0;
                return -1;
        case 0: (*punt_a_buscar)->equil = -1;
                return 0;
        case -1: if (((*punt_a_buscar)->izq)->equil == 1) {
                   dev = rotacion_ID(punt_a_buscar);
                   dev = -1;
                 } /* if */                
                 else
                   dev = rotacion_II(punt_a_buscar);
                 return dev;    
        } /* switch*/       
      } /* if */
      return 0;        
  } /* if */
  else if ((*punt_a_buscar)->clave > a_borrar) {
    dev = borrar_clave_recur(&((*punt_a_buscar)->izq), a_borrar);

    if (dev == -3)
      return dev;
      
    if (dev == -1) {
      switch ((*punt_a_buscar)->equil) {
        case -1: (*punt_a_buscar)->equil = 0;
                 return -1;
        case 0: (*punt_a_buscar)->equil = 1;
                return 0;
        case 1: if (((*punt_a_buscar)->der)->equil == -1) {
                  dev = rotacion_DI(punt_a_buscar);
                  dev = -1;
                } /* if */                                       
                else
                  dev = rotacion_DD(punt_a_buscar);
                return dev;    
        } /* switch*/       
      } /* if */
      return 0;            
  } /* else if */
  else /* ((*punt_a_buscar)->clave == a_borrar) */ {
    dev = -1;
    punt_aux = *punt_a_buscar;
    if ((*punt_a_buscar)->izq == NULL)
      *punt_a_buscar = (*punt_a_buscar)->der;      
    else if ((*punt_a_buscar)->der == NULL)
      *punt_a_buscar = (*punt_a_buscar)->izq;
    else /* ((*punt_a_buscar)->der != NULL) */ /* Buscamos clave menor de las mayores */ {
      dev = reemplazar(&((*punt_a_buscar)->der), &punt_aux);
      free(punt_aux);
         
      if (dev == -1) {
        switch ((*punt_a_buscar)->equil) {
          case 1: (*punt_a_buscar)->equil = 0;
                   return -1;
          case 0: (*punt_a_buscar)->equil = -1;
                  return 0;
          case -1: if (((*punt_a_buscar)->izq)->equil == 1) {
                    dev = rotacion_ID(punt_a_buscar);
                    dev = -1;
                  } /* if */                    
                  else
                    dev = rotacion_II(punt_a_buscar);
                  return dev;    
        } /* switch*/       
      } /* if */
    } /* else */
    return dev;           
  } /* else */
} /* borrar_clave_recur */

void imprime_inorden_inverso(nodo_AVL *punt_recorrer, int margen) {
int i;
  if (punt_recorrer != NULL) {
    imprime_inorden_inverso(punt_recorrer->der, margen+1);
    for (i = 0; i < margen; i++)
      printf("     ");
    printf("%d (%d)\n", punt_recorrer->clave, punt_recorrer->equil);
    imprime_inorden_inverso(punt_recorrer->izq, margen+1);                    
  } /* if ... */    
} /* imprime_inorden */

/* prueba_arbol_AVL.c */

int main(void) {
nodo_AVL *arbol, *punt_encontrado;
tipo_clave clave;

  arbol = NULL;
  while (1) {
    printf("\nIntroduzca clave a insertar (-1 final): ");
    scanf("%d", &clave);
    
    /* printf("Antes de insertar %3d\n", clave); */
    if (clave != -1) 
      insertar_clave_recur(&arbol, clave);
    else
      break;
    /* printf("Despues de insertar %3d\n", clave); */
    
    printf("\n");
    imprime_inorden_inverso(arbol, 0);
    printf("\n");           
  } /* while */
   
  while (1) {
    printf("\nIntroduzca clave a buscar (-1 final): ");
    scanf("%d", &clave);
    
    if (clave != -1)
      punt_encontrado = buscar_clave(arbol, clave);
    else
      break;
    
    printf("\n");
    imprime_inorden_inverso(punt_encontrado, 0);
    printf("\n");           
  } /* while */

  while (1) {
    printf("\nIntroduzca clave a borrar (-1 final): ");
    scanf("%d", &clave);
    
    if (clave != -1)
      borrar_clave_recur(&arbol, clave);
    else
      break;
    
    printf("\n");
    imprime_inorden_inverso(arbol, 0);
    printf("\n");           
  } /* while */
  
  return 0;
} /* main */
